#  sfzen/__init__.py
#
#  Copyright 2024 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Simple object-oriented SFZ parsing and manipulation.
"""
import logging, re
from os import mkdir
from os.path import abspath, basename, dirname, join, splitext, exists
from math import ceil
from operator import or_
from functools import reduce
from collections import defaultdict
from shutil import rmtree
from appdirs import user_cache_dir
from lark import Lark, Transformer, v_args
from lark.tree import Meta
from log_soso import log_error
from sfzen.sfz_elems import (
	Element,
	Modifier,
	Header,
	Global,
	Master,
	Group,
	Region,
	Control,
	Effect,
	Midi,
	Curve,
	Opcode,
	Sample,
	Define,
	Include
)
from sfzen.sort import sorted_opcodes, sorted_opcode_names, sorted_opstrings, midi_note_sort_key

__all__ = [	'COMMENT_DIVIDER',
			'GLOBALIZE_NONE', 'GLOBALIZE_NUMEROUS', 'GLOBALIZE_UNIVERSAL',
			'KEY_OPCODES', 'LOOP_DEFINITION_OPCODES',
			'Control', 'Curve', 'Define', 'Effect', 'Element', 'Global', 'Group', 'Header',
			'Include', 'Master', 'Meta', 'Midi', 'Modifier', 'Opcode', 'Region', 'Sample', 'SFZ',
			'SAMPLES_ABSPATH', 'SAMPLES_COPY', 'SAMPLES_HARDLINK', 'SAMPLES_MOVE',
			'SAMPLES_RESOLVE', 'SAMPLES_SYMLINK', 'SAMPLE_UNIT_OPCODES',
			'midi_note_sort_key', 'sorted_opcode_names', 'sorted_opcodes', 'sorted_opstrings',
			'opcodes_to_dict', 'opstrings_to_dict']

__version__ = "1.8.1"

COMMENT_DIVIDER = '// ' + '-' * 76 + "\n"

SAMPLES_ABSPATH				= 0
SAMPLES_RESOLVE				= 1
SAMPLES_COPY				= 2
SAMPLES_MOVE				= 3
SAMPLES_SYMLINK				= 4
SAMPLES_HARDLINK			= 5

GLOBALIZE_NONE				= 0
GLOBALIZE_NUMEROUS			= 1
GLOBALIZE_UNIVERSAL			= 2

KEY_OPCODES = [
	'lokey',
	'hikey',
	'pitch_keycenter'
]

LOOP_DEFINITION_OPCODES = [
	'egN_loop',
	'egN_loop_count',
	'loop_count',
	'loop_crossfade',
	'loop_end',
	'loop_length_onccN',
	'loop_lengthccN',
	'loop_mode',
	'loop_start',
	'loop_start_onccN',
	'loop_startccN',
	'loop_tune',
	'loop_type',
	'loopcount',
	'loopend',
	'loopmode',
	'loopstart',
	'looptune',
	'looptype'
]

SAMPLE_UNIT_OPCODES = [
	'delay_samples',
	'delay_samples_onccN',
	'end',
	'loop_end',
	'loop_length_onccN',
	'loop_lengthccN',
	'loop_start',
	'loop_start_onccN',
	'loop_startccN',
	'loopend',
	'loopstart',
	'offset',
	'offset_ccN',
	'offset_onccN',
	'offset_random'
]

KEY_RANGE_DICT = {
	'lokey'				: 0,
	'hikey'				: 127
}

VEL_RANGE_DICT = {
	'lovel'				: 0,
	'hivel'				: 127
}

def _list(coll):
	"""
	Convert any basic collection type into a list.
	"""
	if isinstance(coll, list):
		return coll
	if isinstance(coll, dict):
		return list(coll.values())
	return list(coll)

def opcode_dict(coll):
	"""
	Returns a dict {opcode_name:opcode_value} from any basic collection type.
	Elements of the collection may be opstrings or Opcode objects.
	"""
	if len(coll) == 0:
		return []
	coll = _list(coll)
	return opcodes_to_dict(coll) if isinstance(coll[0], Opcode) else opstrings_to_dict(coll)

def opcodes_to_dict(opcode_list: list):
	"""
	Converts a list of Opcode objects into a dict;
	keys are opcode name, values are opcode values.
	"""
	return { opcode.name:opcode.value for opcode in opcode_list }

def opstrings_to_dict(opstring_list: list):
	"""
	Converts a list of string representation (including name and value) into a dict;
	keys are opcode name, values are opcode values.
	"""
	return dict([ opstring.split('=', 1) for opstring in opstring_list ])

def _range_str(rng_dict, opdict):
	d = rng_dict.copy()
	d.update(opdict)
	l = list(d.values())
	return f'{l[0]}-{l[1]}'

def pitch_range_str(opcodes):
	if opcodes:
		opdict = opcode_dict(opcodes)
		return str(opdict['key']) if 'key' in opdict else _range_str(KEY_RANGE_DICT, opdict)
	return '0-127'

def velocity_range_str(opcodes):
	if opcodes:
		opdict = opcode_dict(opcodes)
		return _range_str(VEL_RANGE_DICT, opdict)
	return '0-127'

def range_description(opcodes):
	return f' {pitch_range_str(opcodes)}, {velocity_range_str(opcodes)}'


class Xformer(Transformer):
	"""
	Takes the parse tree generated by lark and generates an SFZ object.
	"""

	header_classes = {
		'global'	: Global,
		'master'	: Master,
		'group'		: Group,
		'region'	: Region,
		'control'	: Control,
		'effect'	: Effect,
		'midi'		: Midi,
		'curve'		: Curve
	}

	def __init__(self, sfz):
		super().__init__()
		self.sfz = sfz
		self.current_header = self.sfz

	@v_args(meta = True)
	def header(self, arg1, arg2):
		"""
		Transformer function which handles <header> tokens.
		"""
		_, toks = self.wonky_lark_args(arg1, arg2)
		header = self.header_classes[toks[0].value]()
		while not self.current_header.may_contain(header):
			self.current_header = self.current_header.parent
		self.current_header.append_subheader(header)
		self.current_header = header

	@v_args(meta = True)
	def define_macro(self, arg1, arg2):
		"""
		Transformer function which handles <define> tokens.
		"""
		meta, toks = self.wonky_lark_args(arg1, arg2)
		try:
			self.sfz.defines[toks[0].value] = Define(toks[0].value, toks[1].value, meta)
		except Exception as e:
			self.sfz.append_parse_error(e, meta)

	@v_args(meta = True)
	def include_macro(self, arg1, arg2):
		"""
		Transformer function which handles <include> tokens.
		"""
		meta, toks = self.wonky_lark_args(arg1, arg2)
		try:
			include = Include(self.unquote(self.replace_defs(toks[0].value)), meta)
			self.sfz.includes.append(include)
			path = join(dirname(self.sfz.filename), include.filename)
			if exists(path):
				try:
					subsfz = SFZ(path, defines = self.sfz.defines, basedir = self.sfz.basedir, is_include = True)
					for header in subsfz.subheaders:
						while not self.current_header.may_contain(header):
							self.current_header = self.current_header.parent
						self.current_header.append_subheader(header)
						self.current_header = header
					self.sfz.defines = subsfz.defines
					self.sfz.includes.extend(subsfz.includes)
				except Exception as e:
					log_error(e)
			else:
				raise ValueError(f'Include not found: "{path}"')
		except Exception as e:
			self.sfz.append_parse_error(e, meta)

	@v_args(meta = True)
	def opcode_exp(self, arg1, arg2):
		"""
		Transformer function which handles <opcode> tokens.
		"""
		meta, toks = self.wonky_lark_args(arg1, arg2)
		try:
			if isinstance(self.current_header, Curve):
				if toks[0] == 'curve_index':
					self.current_header.curve_index = toks[1].value
				else:
					if re.match(r'v(\d+)', toks[0].value):
						self.current_header.points[toks[0].value] = toks[1].value
					else:
						logging.error('Invalid opcode inside velocity curve definition')
			else:
				self.current_header.append_opcode(Opcode(
					self.replace_defs(toks[0].value).lower(),
					self.replace_defs(toks[1].value),
					meta, self.sfz.basedir))
		except Exception as e:
			self.sfz.append_parse_error(e, meta)

	@v_args(meta = True)
	def start(self, arg1, arg2):
		"""
		Transformer function which handles the root of the sfz.
		"""

	def wonky_lark_args(self, arg1, arg2):
		"""
		When @v_args decorates a lark.Transformer function, lark passes parsed tokens
		and meta info to the function. Different versions of lark parser seem to pass
		the tokens and meta arguments in different order. This function fixes that.
		"""
		if isinstance(arg1, Meta):
			return arg1, arg2
		return arg2, arg1

	def replace_defs(self, var):
		"""
		Replace defined variable with the value that it has been defined as in the sfz.
		"""
		return re.sub(r'\$(\w+)', lambda v: self.sfz.defines[v.group(1)].value, var)

	def unquote(self, var):
		"""
		Remove quotes around a parsed value.
		"""
		for q in ["'", '"']:
			if var[0] == q and var[-1] == q:
				return var[1:-1]
		return var


class SFZ(Header):
	"""
	Provides an object-oriented interface to an .sfz file.
	"""

	_parser = None

	def __init__(self, filename = None, *, defines = None, basedir = None, is_include = False):
		"""
		filename: (str) Path to an .sfz file.

		Passing "defines" allows us to construct an SFZ which is an import, and use the
		defined variables from the parent SFZ.

		Passing "basedir" allows included SFZ parts to use the directory of their
		parent when parsing sample paths.
		"""
		super().__init__(None)
		self.filename = filename
		self.defines = defines or {}
		self.is_include = is_include
		self._parent = None
		self.includes = []
		self.parse_errors = []
		if filename is None:
			self.basedir = None
		else:
			self.basedir = dirname(self.filename) if basedir is None else basedir
			if SFZ._parser is None:
				cache_file = join(user_cache_dir(), 'sfzen')
				grammar = join(dirname(__file__), 'res', 'sfz.lark')
				SFZ._parser = Lark.open(grammar, parser = 'lalr',
					propagate_positions = True, cache = cache_file)
			with open(filename) as f:
				tree = SFZ._parser.parse(f.read() + "\n")
			xformer = Xformer(self)
			xformer.transform(tree)

	@property
	def name(self):
		return '[Unnamed]' if self.filename is None else basename(self.filename)

	def may_contain(self, _):
		return True

	def add_region(self):
		"""
		Creates a <region> header and returns it.
		"""
		self.append_subheader(Region())
		return self._subheaders[-1]

	def append_opcode(self, opcode):
		if not self.is_include:
			raise RuntimeError("Opcode outside of header")
		super().append_opcode(opcode)

	def append_parse_error(self, error, meta):
		self.parse_errors.append((error, meta))

	def inherited_opcodes(self):
		return {}

	def global_header(self):
		"""
		Returns the Global header from this SFZ; it one does not exist, creates it at
		the beginning of the header list.
		"""
		for sub in self._subheaders:
			if isinstance(sub, Global):
				return sub
		global_header = Global()
		self._subheaders.insert(0, global_header)
		global_header.parent = self
		return global_header

	def samples(self):
		"""
		Returns all <sample> opcodes contained in this SFZ.
		Generator which yields a Sample object on each iteration.
		"""
		for sub in self._subheaders:
			yield from sub.samples()

	def opcodes_used(self):
		"""
		Returns a set of the keys of all the opcodes used in this SFZ.
		"""
		return reduce(or_, [sub.opcodes_used() for sub in self._subheaders], set())

	def regions_for(self, key = None, lokey = None, hikey = None, lovel = None, hivel = None):
		"""
		Generator function which yields each region contained in this SFZ which matches
		the given criteria.
		For example, to get every region which plays Middle C at any velocity:
			sfz.regions_for(lokey = 60, hikey = 60)
		"""
		for region in self.regions():
			if region.is_triggerd_by(key, lokey, hikey, lovel, hivel):
				yield region

	def save(self):
		"""
		Save to original filename without any further modification.
		"""
		with open(self.filename, 'w', encoding = 'utf-8') as fob:
			self.write(fob)
		return self

	def save_as(self, filename, samples_mode = SAMPLES_ABSPATH, overwrite = True):
		"""
		Save to the given filename.

		"samples_mode" is a constant which defines how to render "sample"
		opcodes. May be one of:

			SAMPLES_ABSPATH		SAMPLES_RESOLVE		SAMPLES_COPY
			SAMPLES_MOVE		SAMPLES_SYMLINK		SAMPLES_HARDLINK

		"""
		filename = abspath(filename)
		target_sfz_dir = dirname(filename)
		filetitle, ext = splitext(basename(filename))
		if not exists(target_sfz_dir):
			mkdir(target_sfz_dir)
		if samples_mode == SAMPLES_ABSPATH:
			for sample in self.samples():
				sample.use_abspath()
		elif samples_mode == SAMPLES_RESOLVE:
			for sample in self.samples():
				sample.resolve_from(target_sfz_dir)
		else:
			samples_path = filetitle + '-samples'
			target_samples_dir = join(target_sfz_dir, samples_path)
			if exists(target_samples_dir):
				if overwrite:
					rmtree(target_samples_dir)
				else:
					raise RuntimeError(f'Sample dir "{target_samples_dir}" already exists')
			mkdir(target_samples_dir)
			for sample in self.samples():
				if samples_mode == SAMPLES_COPY:
					sample.copy_to(target_sfz_dir, samples_path)
				elif samples_mode == SAMPLES_MOVE:
					sample.move_to(target_sfz_dir, samples_path)
				elif samples_mode == SAMPLES_SYMLINK:
					sample.symlink_to(target_sfz_dir, samples_path)
				elif samples_mode == SAMPLES_HARDLINK:
					sample.hardlink_to(target_sfz_dir, samples_path)
		self.filename = filename + '.sfz' if ext == '' else filename
		self.save()

	def write(self, stream):
		"""
		Exports this SFZ to .sfz format.
		"stream" may be any file-like object, including sys.stdout.
		"""
		stream.write(f'{COMMENT_DIVIDER}// {self.name}\n{COMMENT_DIVIDER}\n')
		for sub in self._subheaders:
			sub.write(stream)

	def simplified(self, globalize_mode = GLOBALIZE_NUMEROUS):
		"""
		Returns an equivalent SFZ with common opcodes grouped, and opcodes using the
		default value skipped.
		"""

		simplified_sfz = SFZ()
		simplified_sfz.basedir = self.basedir
		global_header = Global()

		regions = [
			self._clone_sample_region(sample) \
			for sample in self.samples()
		]

		# Condense "lokey", "hikey", "pitch_keycenter"
		for region in regions:
			key_related_values = [
				opcode.value for opcode in region.opcodes.values() \
				if opcode.name in KEY_OPCODES
			]
			if len(key_related_values) == 3 and len(set(key_related_values)) == 1:
				for opcode_name in KEY_OPCODES:
					del region._opcodes[opcode_name]
				region.append_opcode(Opcode('key', key_related_values[0], None))

		# Remove loop-related opcodes for regions that are not looped:
		for region in regions:
			if region.loop_mode or region.loopmode == 'no_loop':
				for opcode_name in LOOP_DEFINITION_OPCODES:
					try:
						del region._opcodes[opcode_name]
					except KeyError:
						pass

		# Place opcodes which have the same value in a majority of regions
		# into the global header:
		if globalize_mode == GLOBALIZE_NUMEROUS and len(regions) > 1:
			opstring_counts = defaultdict(int)
			min_count = ceil(len(regions) / 2)
			for region in regions:
				for opstring in region.opstrings():
					if opstring.split('=', 1)[0] != 'sample':
						opstring_counts[opstring] += 1
			for opstring, count in opstring_counts.items():
				if count >= min_count:
					opcode, value = opstring.split('=', 1)
					for region in regions:
						if opcode in region._opcodes and str(region._opcodes[opcode]) == opstring:
							del region._opcodes[opcode]
					global_header.append_opcode(Opcode(opcode, value))

		# Sort in key order:
		regions.sort(key = midi_note_sort_key)

		# Group regions based on common key:
		key_grouped_regions = defaultdict(list)
		for region in regions:
			key_grouped_regions[midi_note_sort_key(region)].append(region)
		for regions in key_grouped_regions.values():
			if len(regions) > 1:
				group = Group()
				for region in regions:
					group.append_subheader(region)
				common_opstrings = group.common_opstrings()
				if any(len(region._opcodes) > len(common_opstrings) for region in regions):
					for tup in [ opstring.split('=', 1) for opstring in common_opstrings ]:
						group.append_opcode(Opcode(*tup))
						for region in regions:
							del region._opcodes[tup[0]]
					simplified_sfz.append_subheader(group)
				else:
					logging.warning('All region opstrings are common')
					simplified_sfz.append_subheader(regions[0])
			else:
				simplified_sfz.append_subheader(regions[0])

		if globalize_mode == GLOBALIZE_UNIVERSAL and len(list(simplified_sfz.regions())) > 1:
			# Filter global opstrings:
			common_opstrings = simplified_sfz.common_opstrings()
			if len(common_opstrings):
				opstring_tuples = [ opstring.split('=', 1) \
					for opstring in common_opstrings ]
				opstring_tuples = [ tup for tup in opstring_tuples if tup[0] != 'sample' ]
				for tup in opstring_tuples:
					global_header.append_opcode(Opcode(*tup))
				simplified_sfz.remove_opcodes([ tup[0] for tup in opstring_tuples ])

		if len(global_header._opcodes):
			simplified_sfz._subheaders.insert(0, global_header)
			global_header.parent = simplified_sfz

		# Give "Sample" opcodes a basedir for abspath function:
		for elem,_ in simplified_sfz.walk():
			if isinstance(elem, Sample):
				elem.basedir = self.basedir

		return simplified_sfz

	def _clone_sample_region(self, sample):
		region = Region()
		for opcode in sample.parent.inherited_opcodes().values():
			region.append_opcode(opcode)
		return region

	def dump(self):
		"""
		Print (to stdout) a concise outline of this SFZ.
		"""
		for elem, depth in self.walk():
			print('  ' * depth + repr(elem))

	def __repr__(self):
		return f'SFZ {self.filename}'


#  end sfzen/__init__.py



#  sfzen/scripts/sfz_opcode_usage.py
#
#  Copyright 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Utility which lists all opcodes declared in one or many .sfz files
"""
import logging, argparse
from os.path import isdir, isfile, join
from glob import glob
from lark.exceptions import LarkError
from progress.bar import IncrementalBar
from sfzen import SFZ, sorted_opcode_names


def main():
	"""
	Entry point, defined so as to make it easy to reference from bin script.
	"""
	parser = argparse.ArgumentParser()
	parser.epilog = """
	List all the opcodes used by the given SFZ[s].
	"""
	parser.add_argument('Filename', type = str, nargs = '+',
		help = 'File or directory to inspect.')
	parser.add_argument("--recurse", "-r", action = "store_true",
		help = "Recurse into subdirectories.")
	parser.add_argument("--verbose", "-v", action = "store_true",
		help = "Show more detailed debug information.")
	options = parser.parse_args()
	logging.basicConfig(
		level = logging.DEBUG if options.verbose else logging.ERROR,
		format = "[%(filename)24s:%(lineno)3d] %(message)s"
	)

	for path in options.Filename:
		if isdir(path):
			file_list = glob(join(path, '*.sfz'))
			if options.recurse:
				file_list.extend(glob(join(path, '**', '*.sfz')))
		elif isfile(path):
			file_list = [path]
	opcodes = set()
	with IncrementalBar('Reading .sfz', max = len(file_list)) as progress_bar:
		for filename in file_list:
			try:
				opcodes |= SFZ(filename).opcodes_used()
			except LarkError:
				print()
				print(f'Parse error in "{filename}"')
			progress_bar.next()
	print("\n".join(sorted_opcode_names(opcodes)))


if __name__ == "__main__":
	main()


#  end sfzen/scripts/sfz_opcode_usage.py



#  sfzen/sfz_elems.py
#
#  Copyright 2024 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Classes which are instantiated when parsing an .sfz file.
All of these classes are constructed from a lark parser tree Token.
"""
import re, logging
from os import unlink, symlink, link as hardlink, sep as path_separator
from os.path import abspath, exists, join, relpath, realpath
from shutil import move, copy2 as copy
from copy import deepcopy
from functools import cached_property, reduce
try:
	from functools import cache
except ImportError:
	from functools import lru_cache as cache
from operator import and_, or_
from midi_notes import NOTE_PITCHES
from sfzen.sort import sorted_opcodes
from sfzen.opcodes import OPCODES


# ---------------------------
# Elements

class Element:
	"""
	An abstract class which provides parent/child hierarchical relationship.
	This is the base class of all Headers and Opcodes.
	"""

	def __init__(self, meta):
		if meta is None:
			self.line = None
			self.column = None
			self.end_line = None
			self.end_column = None
		else:
			self.line = meta.line
			self.column = meta.column
			self.end_line = meta.end_line
			self.end_column = meta.end_column
		self._parent = None

	@property
	def parent(self):
		"""
		The immediate parent of this element.
		If this is an SFZ, returns None.
		For any other type of element, returns its parent header, or the SFZ if this is
		a top-level header.
		This attribute is set during parsing, and probably shouldn't be modified,
		unless you really know what you are doing.
		"""
		return self._parent

	@parent.setter
	def parent(self, parent):
		self._parent = parent

	def __repr__(self):
		return self.__class__.__name__


class Header(Element):
	"""
	An abstract class which handles the functions common to all SFZ header types.
	Each header type basically acts the same, except for checking what kind of
	subheader it may contain.
	"""

	def __init__(self, meta = None):
		super().__init__(meta)
		self._subheaders = []
		self._opcodes = {}

	def __deepcopy__(self, memo):
		cls = self.__class__
		result = cls.__new__(cls)
		result._subheaders = [
			deepcopy(subheader, memo) \
			for subheader in self._subheaders
		]
		result._opcodes = {
			key:deepcopy(opcode, memo) \
			for key, opcode in self._opcodes.items()
		}
		for subheader in result._subheaders:
			subheader.parent = result
		for opcode in result._opcodes.values():
			opcode.parent = result
		memo[id(self)] = result
		return result

	def may_contain(self, _):
		"""
		This function is used to determine if a header being parsed is a child of the
		last previous header, or the start of an entirely new header group.
		"""
		return False

	def append_opcode(self, opcode):
		"""
		Append an opcode to this Header.

		Returns the appended opcode.
		"""
		self._opcodes[opcode.name] = opcode
		opcode.parent = self
		return opcode

	def append_subheader(self, subheader):
		"""
		Append a subheader to this Header.

		Returns the appended subheader.
		"""
		self._subheaders.append(subheader)
		subheader.parent = self
		return subheader

	@property
	def opcodes(self):
		"""
		Returns a dictionary of Opcode ojects.
		Returns dict { opcode_name:Opcode }
		"""
		return self._opcodes

	@property
	def subheaders(self):
		"""
		Returns a list of headers contained in this Header.
		"""
		return self._subheaders

	def inherited_opcodes(self):
		"""
		Returns all the opcodes defined in this Header with all opcodes defined in its
		parent Header, recursively. Opcodes defined in this Header override parents'.
		Returns dict { opcode_name:Opcode }
		"""
		return self._opcodes if self._parent is None \
			else dict(self._parent.inherited_opcodes(), **self._opcodes)

	def opstrings(self):
		"""
		Returns a set of all the string representation (including name and value) of
		all the opcodes which are used by this Header. This does NOT include opcodes
		used by subheaders beneath this Header.
		"""
		return set(str(opcode) for opcode in self._opcodes.values())

	def opstrings_used(self):
		"""
		Returns a set of all the string representation (including name and value) of
		all the opcodes used in this Header, and any subheaders beneath this Header.
		"""
		opstrings = [sub.opstrings_used() for sub in self._subheaders]
		opstrings.append(self.opstrings())
		return reduce(or_, opstrings, set())

	def common_opstrings(self):
		"""
		Returns a set of all the string representation (including name and value) of
		all the identical opcodes used in every subheader in this Header.
		"""
		if self._subheaders:
			sets = [ sub.common_opstrings() for sub in self._subheaders ]
			# At this point every element of the list is a set of opstrings, one per subheader.
			# Some subheaders have NO common sets, filter these out before reducing to a final set:
			sets = [ set_ for set_ in sets if len(set_) ]
			# Reduce to a single set, or return an empty set if all were empty.
			return reduce(and_, sets) if sets else set()
		return set(str(opcode) for opcode in self.inherited_opcodes().values())

	def uses_opstring(self, opstring):
		"""
		Returns True if the given string representation (including name and value) of
		an opcode is used by this Header. This does not include opcodes used by
		_Headers contained in this Header.
		"""
		return opstring in self.opstrings()

	def opcode(self, name):
		"""
		Returns an Opcode with the given name, if one exists in this Header.
		Returns None if no such opcode exists.
		"""
		return self._opcodes[name] if name in self._opcodes else None

	def iopcode(self, name):
		"""
		Returns an Opcode with the given name, if one exists in this Header or any of
		its ancestors.

		Returns None if no such opcode exists.
		"""
		return self._opcodes[name] if name in self._opcodes \
			else None if self._parent is None \
			else self._parent.iopcode(name)

	def __getattr__(self, name):
		"""
		Returns the value of the opcode with the given "name" contained in this Header,
		or any of its ancestors.

		Raises AttributeError
		"""
		if name[0] == '_':
			return super().__getattribute__(name)
		opcode = self.iopcode(name)
		if opcode:
			return opcode.value
		if normal_opcode(name):
			return None
		raise AttributeError(name)

	def __setattr__(self, name, value):
		"""
		Sets the value of the opcode with the given "name", if it exists.

		Creates the opcode inside this Header, if it does not yet exist.
		"""
		if name[0] == '_':
			super().__setattr__(name, value)
		elif name in self.__dict__:
			self.__dict__[name] = value
		elif '_opcodes' in self.__dict__ and name in self.__dict__['_opcodes']:
			self._opcodes[name].value = value
		elif not normal_opcode(name) is None:
			self.append_opcode(Opcode(name, value, None))
		else:
			super().__setattr__(name, value)

	def opcodes_used(self):
		"""
		Returns a set of the keys of all the opcodes used in this Header and all of
		its subheaders.
		"""
		return set(self._opcodes.keys()) | reduce(or_, [sub.opcodes_used() \
			for sub in self._subheaders], set())

	def regions(self):
		"""
		Returns all <region> headers contained in this Header and all of its
		subheaders.
		This is a generator function which yields a Region object on each iteration.
		"""
		for sub in self._subheaders:
			if isinstance(sub, Region):
				yield sub
			yield from sub.regions()

	def samples(self):
		"""
		This is a generator function which yields a Sample object on each iteration.
		"""
		if 'sample' in self._opcodes:
			yield self._opcodes['sample']
		for sub in self._subheaders:
			yield from sub.samples()

	def walk(self, depth = 0):
		"""
		Generator which recusively yields every element contained in this Header,
		including opcodes and subheaders. Opcodes are yielded first, then subheaders.
		Each iteration returns a tuple (Element, (int) depth)
		"""
		yield (self, depth)
		depth += 1
		for opcode in self._opcodes.values():
			yield (opcode, depth)
		for sub in self._subheaders:
			yield from sub.walk(depth)

	def opcode_count(self):
		"""
		Returns (int) number of opcodes used in this Header and all subheaders
		"""
		return sum(len(elem.opcodes.values()) \
			for elem, _ in self.walk() \
			if isinstance(elem, Header))

	def reduce_common_opcodes(self):
		"""
		Move common opcodes (name/value) from contained headers to this header.
		"""
		if self._subheaders:
			common_opstrings = self.common_opstrings()
			for tup in [ opstring.split('=', 1) for opstring in common_opstrings ]:
				self.append_opcode(Opcode(*tup))
				for sub in self._subheaders:
					del sub._opcodes[tup[0]]

	def remove_opcodes(self, opcode_list):
		for elem, _ in self.walk():
			if isinstance(elem, Header):
				elem._opcodes = { key:opcode \
					for key, opcode in elem._opcodes.items() \
					if key not in opcode_list }

	def __str__(self):
		return f'<{self.__class__.__name__.lower()}>'

	def __repr__(self):
		return f'{self.__class__.__name__} ({len(self._opcodes):d} opcodes)'

	def write(self, stream):
		"""
		Exports this Header and all of it's contained headers and
		opcodes to .sfz format.
		"stream" may be any file-like object, like "sys.stdout".
		"""
		stream.write(str(self) + "\n")
		if self._opcodes:
			for op in sorted_opcodes(self._opcodes.values()):
				op.write(stream)
			stream.write("\n")
		if self._subheaders:
			for sub in self._subheaders:
				sub.write(stream)


class Modifier(Element):
	"""
	Abstract class which is inherited by Define & Include.
	"""


class Global(Header):
	"""
	Represents an SFZ Global header. Created by Lark transformer when importing SFZ.
	"""

	def may_contain(self, _):
		return True


class Master(Header):
	"""
	Represents an SFZ Master header. Created by Lark transformer when importing SFZ.
	"""

	def may_contain(self, header):
		return type(header) not in [Global, Master]


class Group(Header):
	"""
	Represents an SFZ Group header. Created by Lark transformer when importing SFZ.
	"""

	def may_contain(self, header):
		return type(header) not in [Global, Master, Group]


class Region(Header):
	"""
	Represents an SFZ Region header. Created by Lark transformer when importing SFZ.
	"""

	def may_contain(self, header):
		"""
		Used during parsing to determine where to append a newly parsed Header
		"""
		return type(header) not in [Global, Master, Group, Region]

	def is_triggerd_by(self, key = None, lokey = None, hikey = None, lovel = None, hivel = None):
		"""
		Returns boolean True/False if this Region matches the given criteria.
		For example, to test if this region plays Middle C at any velocity:
			region.is_triggerd_by(lokey = 60, hikey = 60)
		"""
		if key is None and lokey is None and hikey is None and lovel is None and hivel is None:
			raise RuntimeError('Requires a key or velocity to test against')
		ops = self.inherited_opcodes()
		if key is not None and 'key' in ops and ops['key'].value != key:
			return False
		if lokey is not None and 'lokey' in ops and ops['lokey'].value > lokey:
			return False
		if hikey is not None and 'hikey' in ops and ops['hikey'].value < hikey:
			return False
		if lovel is not None and 'lovel' in ops and ops['lovel'].value > lovel:
			return False
		if hivel is not None and 'hivel' in ops and ops['hivel'].value < hivel:
			return False
		return True

	@property
	def sample(self):
		"""
		Returns the Sample opcode contained in this <region>
		Returns None if no Sample has yet defined.
		"""
		return self.opcode('sample')


class Control(Header):
	"""
	Represents an SFZ Control header. Created by Lark transformer when importing SFZ.
	"""


class Effect(Header):
	"""
	Represents an SFZ Effect header. Created by Lark transformer when importing SFZ.
	"""


class Midi(Header):
	"""
	Represents an SFZ MIDI header. Created by Lark transformer when importing SFZ.
	"""


class Curve(Header):
	"""
	Represents an SFZ curve. Created by Lark transformer when importing SFZ.
	"""

	curve_index = None
	points = {}

	def __str__(self):
		return f'<curve>curve_index={self.curve_index}'

	def write(self, stream):
		"""
		Exports this Curve to .sfz format.
		"stream" may be any file-like object, including sys.stdout.
		"""
		stream.write(str(self) + "\n")
		for key, val in self.points.items():
			stream.write(f'{key}={val}\n')


class Opcode(Element):
	"""
	Represents an SFZ opcode. Created by Lark transformer when importing SFZ.
	"""

	def __new__(cls, name, value, meta = None, basedir = None):
		return super().__new__(Sample) if name == 'sample' else super().__new__(Opcode)

	def __init__(self, name, value, meta = None, _ = None):
		super().__init__(meta)
		self.name = name
		self.value = value
		self.basedir = None

	def __deepcopy__(self, memo):
		cls = self.__class__
		result = cls.__new__(cls, self.name, self.value)
		result.name = self.name
		if cls is Sample:
			result.basedir = self.basedir
			result._value = self.abspath
		else:
			result._value = self._value
		memo[id(self)] = result
		return result

	@property
	def value(self):
		"""
		Returns the value as the type defined in the opcode definition.
		"""
		return self._value

	@value.setter
	def value(self, value):
		"""
		Set the value of this Opcode, after converting the given "value" to
		the data type of this Opcode, (see "type()").
		"""
		if self.type is float:
			self._value = float(value)
		elif self.type == int:
			try:
				self._value = int(value)
			except ValueError as err:
				if value.upper() in NOTE_PITCHES:
					self._value = NOTE_PITCHES[value.upper()]
				else:
					raise err
		else:
			self._value = value

	@property
	def abspath(self):
		"""
		Not implemented in Opcode, but in derived class Sample.
		"""
		raise NotImplementedError()

	@cached_property
	def type(self):
		"""
		Returns the data type (i.e. float, int, or str) of this Opcode.
		"""
		return data_type(self.name)

	@cached_property
	def type_str(self):
		"""
		Returns the string "type" defined in the opcode definition.
		"""
		return self._def_value('type')

	@cached_property
	def unit(self):
		"""
		Returns the unit defined in the opcode definition.
		"""
		return self._def_value('unit')

	@cached_property
	def validation_rule(self):
		"""
		Returns the validation rule defined in the opcode definition.
		"""
		return self._def_value('valid')

	@cached_property
	def validator(self):
		"""
		Returns a class which extends Validator
		"""
		return validator_for(self.name)

	@cached_property
	def definition(self):
		"""
		Returns the defintion of this opcode from the SFZ syntax (see opcodes.py)
		The defintion name is normalized, replacing "_ccN" -type elements.
		"""
		return opcode_definition(self.name)

	def _def_value(self, key):
		"""
		Returns the attribute of the opcode defintion specified by the given "key".
		If there is no opcode definition found, returns None.
		"""
		return None \
			if self.definition is None or 'value' not in self.definition \
			else self.definition['value'][key]

	def __str__(self):
		return '%s=%s' % (self.name, self._value)

	def __repr__(self):
		return f'{self.__class__} {self}'

	def write(self, stream):
		"""
		Exports this Opcode to .sfz format.
		"stream" may be any file-like object, (including sys.stdout).
		"""
		stream.write(str(self) + "\n")


class Sample(Opcode):
	"""
	Unique case Opcode with extra functions for path manipulation.
	"""

	RE_PATH_DIVIDER = '[\\\/]'

	def __init__(self, name, value, meta = None, basedir = None):
		"""
		When instantiating a Sample, the "name" is always "Sample"; the "value"
		is the path to the sample as it appears in the SFZ. In Opcode.__init__(),
		the internal "_value" property is set to the given value.

		The "meta" argument is passed to the constructor during parsing of an existing
		SFZ file. You can safely ignore it when constructing an SFZ from scratch.

		"basedir" would be the name of the directory containing the SFZ which owns this
		Sample. This value is used for determing the relative path to a sample defined
		with an absolute path.
		"""
		super().__init__(name, value, meta)
		self.basedir = basedir

	@property
	def _path_parts(self):
		"""
		Splits the directory / filenames of the parsed value of this opcode
		Returns list of str
		"""
		return re.split(self.RE_PATH_DIVIDER, self._value)

	@property
	def abspath(self):
		"""
		Returns (str) the absolute path to the sample
		"""
		path = path_separator + join(*self._path_parts)
		return path if exists(path) else abspath(join(self.basedir, *self._path_parts))

	@property
	def relpath(self):
		"""
		Returns (str) the path to the sample relative to the parent SFZ
		"""
		return join(*self._path_parts)

	@property
	def basename(self):
		"""
		Returns (str) the basename of the sample
		"""
		return self._path_parts[-1]

	def exists(self):
		"""
		Returns boolean True if file exists
		"""
		return exists(self.abspath)

	def use_abspath(self):
		"""
		Directs this Sample to use an absolute path when writing .sfz
		"""
		self._value = self.abspath

	def _new_target(self, sfz_directory, samples_path):
		"""
		Returns a tuple ((str) value, (str) abspath),
		where "value" is the new opcode value for this Sample, and
		"abspath" is the target of a copy/move/symlink/hardlink op.
		"""
		return (
			join(samples_path, self.basename),
			join(sfz_directory, samples_path, self.basename)
		)

	def resolve_from(self, sfz_directory):
		"""
		Directs this Sample to use a relative path when writing .sfz.

		"sfz_directory" is the directory in which the .sfz file is to be written.
		"""
		self._value = relpath(realpath(self.abspath), realpath(sfz_directory))

	def copy_to(self, sfz_directory, samples_path):
		"""
		Copies the source sample to a new location and sets the value of this "sample"
		opcode to point to the new location.

		"sfz_directory" is the directory in which the .sfz file is to be written.

		"samples_path" is the directory where the samples are to be written, relative
		to "sfz_directory".
		"""
		value, path = self._new_target(sfz_directory, samples_path)
		if exists(path):
			unlink(path)
		copy(self.abspath, path)
		self._value = value

	def move_to(self, sfz_directory, samples_path):
		"""
		Moves the source sample to a new location and sets the value of this "sample"
		opcode to point to the new location.

		"sfz_directory" is the directory in which the .sfz file is to be written.

		"samples_path" is the directory where the samples are to be written, relative
		to "sfz_directory".
		"""
		value, path = self._new_target(sfz_directory, samples_path)
		if exists(path):
			unlink(path)
		move(self.abspath, path)
		self._value = value

	def symlink_to(self, sfz_directory, samples_path):
		"""
		Symlinks the source sample in a new samples directory and sets the value of
		this "sample" opcode to point to the new location.

		"sfz_directory" is the directory in which the .sfz file is to be written.

		"samples_path" is the directory where the samples are to be written, relative
		to "sfz_directory".
		"""
		value, path = self._new_target(sfz_directory, samples_path)
		if exists(path):
			unlink(path)
		symlink(realpath(self.abspath), realpath(path))
		self._value = value

	def hardlink_to(self, sfz_directory, samples_path):
		"""
		Hard links the source sample in a new samples directory and sets the value of
		this "sample" opcode to point to the new location.

		"sfz_directory" is the directory in which the .sfz file is to be written.

		"samples_path" is the directory where the samples are to be written, relative
		to "sfz_directory".
		"""
		value, path = self._new_target(sfz_directory, samples_path)
		if exists(path):
			unlink(path)
		hardlink(self.abspath, path)
		self._value = value


class Define(Modifier):
	"""
	Represents a Define Opcode. Created by Lark transformer when importing SFZ.
	"""

	def __init__(self, varname, value, meta):
		super().__init__(meta)
		self.varname = varname
		self.value = value


class Include(Modifier):
	"""
	Represents an Include Opcode. Created by Lark transformer when importing SFZ.
	"""

	def __init__(self, filename, meta):
		super().__init__(meta)
		self.filename = filename


# ---------------------------
# Validators

class Validator:
	"""
	Abstract class which is the base for the various validator classes.
	"""

	type = None

	def type_name(self):
		return "any" if self.type is None else self.type.__name__


class AnyValidator(Validator):
	"""
	Validates an opcode when any value is valid.
	"""

	def is_valid(self, *_):
		return True


class ChoiceValidator(Validator):
	"""
	Validates an opcode when valid values are in a predefined list of choices.
	"""

	@classmethod
	def from_rule(cls, str_choices, type_):
		return ChoiceValidator(
			[ c.strip("' []") for c in str_choices.split(',') ],
			type_)

	def __init__(self, choices, type_):
		self.choices = choices
		self.type = type_

	def is_valid(self, value, validate_type = True):
		if validate_type and not isinstance(value, self.type):
			return False
		return value in self.choices


class RangeValidator(Validator):
	"""
	Validates an opcode when valid values are within a range.
	"""

	@classmethod
	def from_rule(cls, rulestr, type_):
		lo, hi = rulestr.split(',')
		if type_ is None:
			type_ = int
		return RangeValidator(type_(lo), type_(hi), type_)

	def __init__(self, lowval, highval, type_):
		self.lowval = lowval
		self.highval = highval
		self.type = type_

	def is_valid(self, value, validate_type = True):
		if validate_type and not isinstance(value, self.type):
			return False
		return self.lowval <= value <= self.highval


class MinValidator(Validator):
	"""
	Validates an opcode when valid values are from a set minimum to any upper value.
	"""

	@classmethod
	def from_rule(cls, rulestr, type_):
		if type_ is None:
			type_ = int
		return MinValidator(type_(rulestr), type_)

	def __init__(self, lowval, type_):
		self.lowval = lowval
		self.highval = None
		self.type = type_

	def is_valid(self, value, validate_type = True):
		if validate_type and not isinstance(value, self.type):
			return False
		return self.lowval <= value


@cache
def validator_for(opcode_name):
	"""
	Returns a class which extends Validator
	"""
	rule = validation_rule(opcode_name)
	if rule is None:
		return AnyValidator()
	match = re.match(r'^(Choice|Range|Min|Any)\(([^\)]*)\)', rule)
	if match is None:
		raise RuntimeError(f'Invalid validation rule: "{rule}"')
	type_ = data_type(opcode_name)
	if match.group(1) == 'Choice':
		return ChoiceValidator.from_rule(match.group(2), type_)
	if match.group(1) == 'Range':
		return RangeValidator.from_rule(match.group(2), type_)
	if match.group(1) == 'Min':
		return MinValidator.from_rule(match.group(2), type_)
	return AnyValidator()

@cache
def validation_rule(opcode_name):
	definition = opcode_definition(opcode_name)
	if definition is None:
		return None
	try:
		rule = definition["value"]["valid"]
	except KeyError:
		return validation_rule(definition["modulates"]) \
			if "modulates" in definition else None
	match = re.match(r'^(Any|Alias|Choice|Range|Min)\(([^\)]*)\)', rule)
	if match is None:
		raise RuntimeError('Invalid validation rule: ' + rule)
	return validation_rule(match.group(2).strip("'")) \
		if match.group(1) == 'Alias' \
		else match.group(0)

@cache
def data_type(opcode_name):
	"""
	Normalizes an opcode_name and returns the data type.
	"""
	definition = opcode_definition(opcode_name)
	if definition is None:
		return None
	if "value" not in definition or "type" not in definition["value"]:
		return data_type(definition["modulates"]) \
			if "modulates" in definition else None
	if definition["value"]["type"] == 'float':
		return float
	if definition["value"]["type"] == 'integer':
		return int
	if definition["value"]["type"] == 'string':
		return str
	raise TypeError("unknown type: " + definition["value"]["type"])

@cache
def modulates(opcode_name):
	"""
	Returns the name of the opcode that the given opcode modulates, if applicable.
	"""
	definition = opcode_definition(opcode_name)
	try:
		return definition["modulates"]
	except KeyError:
		return None

@cache
def opcode_definition(opcode_name):
	"""
	Normalizes an opcode_name and returns the matching opcode definition.
	"""
	opcode_name = normal_opcode(opcode_name)
	return None if opcode_name is None else OPCODES[opcode_name]

@cache
def normal_opcode(opcode_name, follow_aliases = True):
	"""
	Normalizes a "_ccN" opcode opcode_name.
	If "follow_aliases" is True, returns the name of the opcode that this opcode aliases.
	"""
	if opcode_name is None:
		logging.warning('opcode_name is None')
		return None
	if opcode_name in OPCODES:
		return aliases(opcode_name) if follow_aliases else opcode_name
	if re.match(r'amp_velcurve_(\d+)', opcode_name):
		return 'amp_velcurve_N'
	if re.search(r'eq\d+_', opcode_name):
		opcode_name = re.sub(r'eq\d+_', 'eqN_', opcode_name)
		if opcode_name in OPCODES:
			return aliases(opcode_name) if follow_aliases else opcode_name
		if re.search(r'cc\d', opcode_name):
			for regex, repl in [
				(r'_oncc(\d+)', '_onccX'),
				(r'_cc(\d+)', '_ccX'),
				(r'cc(\d+)', 'ccX')
			]:
				sub = re.sub(regex, repl, opcode_name)
				if sub != opcode_name and sub in OPCODES:
					return aliases(sub) if follow_aliases else opcode_name
	if re.search(r'cc\d', opcode_name):
		for regex, repl in [
			(r'_oncc(\d+)', '_onccN'),
			(r'_cc(\d+)', '_ccN'),
			(r'cc(\d+)', 'ccN')
		]:
			sub = re.sub(regex, repl, opcode_name)
			if sub != opcode_name:
				# Recurse for opcodes like "eq3_gain_oncc12"
				if sub in OPCODES:
					return aliases(sub) if follow_aliases else opcode_name
				logging.debug('normal_opcode: Falling through to do recursive checking')
				return normal_opcode(sub, follow_aliases)
	return None

@cache
def aliases(opcode_name, only_alias = False):
	"""
	Returns the opcode which the given opcode aliases, (if it does).
	If it is not aliasing another opcode, the return value depends upon the
	"only_alias" argument. When "only_alias" is True, and the given opcode does not
	alias another opcode, returns None. When "only_alias" is False (the default),
	and the given opcode does not alias another opcode, returns the given opcode.
	"""
	definition = OPCODES[opcode_name]
	if definition is not None:
		try:
			match = re.match(r'Alias\([\'"](\w+)[\'"]\)', definition['value']['valid'])
		except KeyError:
			pass
		else:
			if match:
				return match.group(1)
	return None if only_alias else opcode_name


#  end sfzen/sfz_elems.py



#  sfzen/resampler.py
#
#  Copyright 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides class used to modify sample rate, bit depth, and number of channels of SFZ samples.
"""
import logging
from os.path import join, basename
from tempfile import gettempdir as tempdir
try:
	import sox
except ImportError:
	print("""
Oops! "sox" is not installed! Resampling depends upon it
(You may have installed sfzen without the [resample] option)
Install the "sox" python wrapper and try again:

   pip install sox
""")
	exit()
from sfzen import SAMPLE_UNIT_OPCODES, SAMPLES_MOVE
from sfzen.sfz_elems import Opcode


class SFZResampler:
	"""
	Class used to modify sample rate, bit depth, and number of channels of SFZ samples.
	"""

	def __init__(self, sfz, target_rate = 44100, target_mono = False, target_bitdepth = 16):
		self.sfz = sfz
		self.target_rate = target_rate
		self.target_mono = target_mono
		self.target_bitdepth = target_bitdepth
		self.resamplers = [
			SampleResampler(sample, self.target_rate, self.target_mono, self.target_bitdepth) \
			for sample in self.sfz.samples() ]
		self.bad_samplers = [ resampler for resampler in self.resamplers if resampler.needs_resample() ]

	def needs_resample(self):
		return bool(self.bad_samplers)

	def resample_as(self, filename):
		for resampler in self.resamplers:
			if resampler.needs_resample():
				resampler.resample()
		self.sfz.save_as(filename, SAMPLES_MOVE)


class SampleResampler:

	soxi_infos = {}			# { abspath : sox.file_info }
	converted_files = []	# Complicated filename in temp dir.

	def __init__(self, sample, target_rate, target_mono, target_bitdepth):
		self.sample = sample
		self.basedir = sample.basedir
		assert self.basedir is not None
		self.target_rate = target_rate
		self.target_mono = target_mono
		self.target_bitdepth = target_bitdepth
		self.source_path = self.sample.abspath
		if not self.source_path in self.soxi_infos:
			self.soxi_infos[self.source_path] = sox.file_info.info(self.source_path)
		for key, value in self.soxi_infos[self.source_path].items():
			setattr(self, key, value)

	@property
	def offset(self):
		return self.sample.parent.offset or 0

	@property
	def loop_start(self):
		return self.sample.parent.loop_start or self.sample.parent.loopstart or 0

	@property
	def loop_end(self):
		return self.sample.parent.loop_end or self.sample.parent.loopend or self.num_samples

	def needs_resample(self):
		return self.sample_rate != self.target_rate \
			or self.target_mono and self.channels > 1 \
			or self.bitdepth != self.target_bitdepth

	def resample(self):
		if not self.needs_resample():
			logging.debug('Not resampling %s - already in target format', basename(self.sample.path))
			return
		# -----------------------------------------------------
		# Rescale other opcodes affected if sample rate changed
		if self.sample_rate != self.target_rate:
			ratio = self.target_rate / self.sample_rate
			for opcode_name in SAMPLE_UNIT_OPCODES:
				opcode = self.sample.parent.iopcode(opcode_name)	# Retrieves inherited as well
				if opcode:
					adjusted_value = round(float(opcode.value) * ratio)
					if adjusted_value != opcode.value:
						self.sample.parent._opcodes[opcode_name] = Opcode(
							opcode_name, adjusted_value, None, self.basedir)
						logging.debug('Adjusted %s', opcode)
		# ------------------
		# Do file conversion
		out_file = join(tempdir(), basename(self.source_path))
		if not out_file in self.converted_files:
			xfmr = sox.Transformer()
			xfmr.convert(
				samplerate = self.target_rate,
				n_channels = 1 if self.target_mono else None,
				bitdepth = self.target_bitdepth)
			xfmr.build_file(self.source_path, out_file)
			self.converted_files.append(out_file)
		self.sample.value = out_file


#  end sfzen/resampler.py



#  sfzen/sort.py
#
#  Copyright 2024 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Provides functions used for sorting opcodes in a preferred order.
"""

OPCODE_SORT_ORDER = [
	'lokey',
	'hikey',
	'key',
	'lovel',
	'hivel',
	'lochan',
	'hichan',
	'sample',
	'pitch_keycenter',
	'loop_mode',
	'loop_start',
	'loop_end',
	'offset',
	'group',
	'off_by',
	'ampeg_attack',
	'ampeg_decay',
	'ampeg_delay',
	'ampeg_hold',
	'ampeg_release',
	'ampeg_sustain',
	'amplfo_delay',
	'amplfo_depth',
	'amplfo_freq',
	'volume',
	'pan',
	'cutoff',
	'resonance',
	'transpose',
	'tune',
	'pitch_keytrack',
	'pitch_veltrack',
	'fileg_delay',
	'fileg_attack',
	'fileg_decay',
	'fileg_depth',
	'fileg_sustain',
	'fileg_release',
	'fileg_hold',
	'fil_type',
	'fil_veltrack',
	'fillfo_delay',
	'fillfo_depth',
	'fillfo_freq',
	'effect1',
	'effect2',
	'pitcheg_delay',
	'pitcheg_attack',
	'pitcheg_decay',
	'pitcheg_depth',
	'pitcheg_sustain',
	'pitcheg_release',
	'pitcheg_hold',
	'pitchlfo_delay',
	'pitchlfo_depth',
	'pitchlfo_freq'
]

def sorted_opcodes(opcodes):
	"""
	Sort a list of Opcode objects according to preferred OPCODE_SORT_ORDER.
	"""
	return sorted(opcodes, key = lambda opcode: \
		OPCODE_SORT_ORDER.index(opcode.name) if opcode.name in OPCODE_SORT_ORDER else 1000)

def sorted_opcode_names(opcodes):
	"""
	Sort a list of strings (opcode names), according to preferred OPCODE_SORT_ORDER.
	"""
	return sorted(opcodes, key = lambda opcode: \
		OPCODE_SORT_ORDER.index(opcode) if opcode in OPCODE_SORT_ORDER else 1000)

def sorted_opstrings(opstrings):
	def sort_val(opstring):
		op = opstring.split('=', 1)[0]
		return OPCODE_SORT_ORDER.index(op) if op in OPCODE_SORT_ORDER else 1000
	return sorted(opstrings, key = sort_val)

def midi_note_sort_key(region):
	"""
	Provides a key to use for sorting a list of Regions based on "lokey", "hikey" values.
	"""
	key = region.key
	if key is None:
		lokey = region.lokey or 1
		hikey = region.hikey or 127
	else:
		lokey = key
		hikey = key
	return lokey * 128 + hikey


#  end sfzen/sort.py



#  sfzen/utils/sample_unit_opcodes.py
#
#  Copyright 2025 Leon Dionne <ldionne@dridesign.sh.cn>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#
"""
Compile a list of opcodes which use "sample units"
"""
from pretty_repr import Repr
from sfzen.opcodes import OPCODES
from sfzen.sfz_elems import aliases, modulates

SAMPLE_UNIT_OPCODES = []
for opcode in OPCODES.values():
	try:
		unit = opcode['value']['unit']
	except KeyError:
		continue
	if unit == 'sample units':
		SAMPLE_UNIT_OPCODES.append(opcode['name'])

for opcode in OPCODES.values():
	alias = aliases(opcode['name'], True)
	if alias in SAMPLE_UNIT_OPCODES:
		SAMPLE_UNIT_OPCODES.append(opcode['name'])

for opcode in OPCODES.values():
	mod = modulates(opcode['name'])
	if mod and mod in SAMPLE_UNIT_OPCODES:
		SAMPLE_UNIT_OPCODES.append(opcode['name'])

SAMPLE_UNIT_OPCODES = list(set(SAMPLE_UNIT_OPCODES))
SAMPLE_UNIT_OPCODES.sort()
Repr(SAMPLE_UNIT_OPCODES).print()

#  end sfzen/utils/sample_unit_opcodes.py
